# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL

"""Convenience classes and functions that allow you to quickly build gp apps."""

import os
import sys
import time
import inspect
import random
import logging
import argparse
import operator
from contextlib import contextmanager

import dill
import numpy as np
import toolz
import deap
import deap.tools

from . import gp
from . import utils
from .assessment import SingleProcessFactoy


class GPRunner(object):
    """Runner for gp problem sets.

    Takes care of propper initialization, execution, and accounting of a gp run
    (i.e. population creation, random state, generation count, hall of fame, and
    logbook). The method init() has to be called once before stepping through
    the evolution process with method step(); init() and step() invoke the
    assessment runner.
    """

    def __init__(self, IndividualClass, algorithm_factory, assessment_runner, callbacks=()):
        """Init GPRunner.

        :param IndividualClass: Class inherited from gp.AExpressionTree.
        :param algorithm_factory: callable() -> gp algorithm, as defined in
                                   gp.algorithms.
        :param assessment_runner: callable(population) -> None, updates fitness
                                   values of each invalid individual in population.
        """
        self.IndividualClass = IndividualClass
        self.algorithm_factory = algorithm_factory
        self.assessment_runner = assessment_runner
        self.halloffame = []
        self.logbook = ''
        self.callbacks = callbacks

    def init(self, pop_size):
        """Initialize the gp run."""
        self.halloffame = deap.tools.ParetoFront()
        self.logbook = deap.tools.Logbook()
        self.mstats = None
        self.step_count = 0
        with random_state(self):
            self.population = self.IndividualClass.create_population(pop_size)

        self.algorithm = self.algorithm_factory()
        self._update()

    def step(self):
        """Step through the evolution process."""
        with random_state(self):
            self.population = self.algorithm.evolve(self.population)
        self.step_count += 1
        self._update()
        for cb in self.callbacks:
            cb(self)

    def _update(self):
        """Evaluate invalid individuals and update statistics accordingly."""
        evals = self.assessment_runner(self.population)
        self.halloffame.update(self.population)
        if not self.mstats:
            self.mstats = create_stats(len(self.population[0].fitness.values))
        record = self.mstats.compile(self.population)
        self.logbook.record(gen=self.step_count, evals=evals, **record)



@contextmanager
def random_state(cls, rng=random):
    """Do work inside this contextmanager with a random state defined by cls.

    Looks for _prev_state to seed the rng.
    On exit, it will write the current state of the rng as _tmp_state
    to the cls.

    :params cls: Any object.
    :params rng: Instance of a random number generator.
    """
    cls._tmp_state = rng.getstate()
    rng.setstate(getattr(cls, '_prev_state', rng.getstate()))
    yield
    cls._prev_state = rng.getstate()
    rng.setstate(getattr(cls, '_tmp_state', rng.getstate()))


def default_gprunner(Individual, assessment_runner, callbacks=(), **kwargs):
    """Create a default GPRunner instance.

    For config options see `MateFactory`, `MutateFactory`, `AlgorithmFactory`.
    """
    default_config = dict(mating='cxonepoint', mating_max_height=20,
                          mutation='mutuniform', mutation_max_height=20,
                          algorithm='nsga2', crossover_prob=0.5,
                          mutation_prob=0.2, tournament_size=2,
                          select='nsga2', create_method='halfandhalf',
                          create_min_height=1, create_max_height=4,
                          mutate_tree_max=2, mutate_tree_min=0)

    default_config.update(kwargs)
    mate = MateFactory.create(default_config, Individual)
    mutate = MutateFactory.create(default_config, Individual)
    select = SelectFactory.create(default_config)
    create_method = CreateFactory.create(default_config, Individual)
    AlgorithmFactory.create(default_config, mate, mutate, select, create_method)  # A test run to check config params.
    algorithm_factory = toolz.partial(AlgorithmFactory.create, default_config, mate, mutate, select, create_method)
    return GPRunner(Individual, algorithm_factory, assessment_runner, callbacks=callbacks)


def make_checkpoint(app):
    if app.valid_checkpointing and (app.gp_runner.step_count % app.args.checkpoint_frequency == 0):
        app.checkpoint()


DEFAULT_CALLBACKS = make_checkpoint,


class Application(object):
    """An application based on `GPRunner`.

    Controls execution of the runner and adds checkpointing and logging
    functionality; also defines a set of available command line options and
    their default values.

    To create a full console application one can use the factory function
    create_console_app().
    """

    def __init__(self, config, gp_runner, checkpoint_file, callbacks=()):
        """
        :param config: Container holding all configs
        :type config: dict or argparse.Namespace
        :param gp_runner: Instance of `GPRunner`
        :param checkpoint_file: Path to checkpoint_file
        :param callbacks:
        """
        self.args = to_argparse_namespace(config)
        self.gp_runner = gp_runner
        self.checkpoint_file = checkpoint_file
        self.pareto_fronts = []
        self.logger = logging.getLogger(__name__)
        self._initialized = False
        self.callbacks = callbacks + DEFAULT_CALLBACKS

    @property
    def assessment_runner(self):
        return self.gp_runner.assessment_runner

    def run(self, break_condition=None):
        """Run gp app.

        :param break_condition: is called after every evolutionary step.
        :type break_condition: callable(application)

        :return: number of iterations executed during run.
        """
        if break_condition is None:
            break_condition = lambda app: False
        iterations = 0
        if self.args.pop_size < 1:
            return iterations
        if not self._initialized:
            random.seed(self.args.seed)
            self.gp_runner.init(self.args.pop_size)
            self._update()
            iterations += 1
        while self.gp_runner.step_count < self.args.num_generations and not break_condition(self):
            self.gp_runner.step()
            self._update()
            iterations += 1
        return iterations

    def _update(self):
        for cb in self.callbacks:
            cb(self)

    @property
    def valid_checkpointing(self):
        return self.checkpoint_file is not None and isinstance(self.args.checkpoint_frequency, int)

    def checkpoint(self):
        """Checkpoint current state of evolution."""
        safe(self.checkpoint_file, args=self.args, runner=self.gp_runner,
             random_state=random.getstate(), pareto_fronts=self.pareto_fronts, callbacks=self.callbacks)
        self.logger.debug('Saved checkpoint to {}'.format(self.checkpoint_file))

    @property
    def workdir(self):
        return os.path.dirname(os.path.abspath(self.checkpoint_file))

    @classmethod
    def from_checkpoint(cls, file_name):
        """Create application from checkpoint file."""
        cp = load(file_name)
        app = cls(cp['args'], cp['runner'], file_name, callbacks=cp['callbacks'])
        app.pareto_fronts = cp['pareto_fronts']
        app._initialized = True
        random.setstate(cp['random_state'])
        return app

    @staticmethod
    def add_options(parser):
        """Add available parser options."""
        parser.add_argument('--pop_size', '-p', dest='pop_size', metavar='n',
                            type=utils.argparse.non_negative_int, default=10,
                            help='initial population size (default: 10)')
        parser.add_argument('--num_generations', '-n', dest='num_generations', metavar='n',
                            type=utils.argparse.non_negative_int, default=10,
                            help='number of generations to evolve (default: 10)')
        parser.add_argument('--seed', dest='seed', metavar='n', type=utils.argparse.non_negative_int,
                            default=random.randint(0, sys.maxsize),
                            help='a seed for the random genrator (default: random.randint(0, sys.maxsize))')
        parser.add_argument('--checkpoint_frequency', '-f', dest='checkpoint_frequency', metavar='n',
                            type=utils.argparse.positive_int, default=1,
                            help='do checkpointing every n generations (default: 1)')


def default_console_app(IndividualClass, AssessmentRunnerClass, parser=argparse.ArgumentParser(), callbacks=()):
    """Factory function for a console application."""
    Application.add_options(parser)
    cp_group = parser.add_mutually_exclusive_group(required=False)
    cp_group.add_argument('--resume', dest='resume_file', metavar='FILE', type=str,
                          help='continue previous run from a checkpoint file')
    cp_group.add_argument('-o', dest='checkpoint_file', metavar='FILE', type=str,
                          default=os.path.join('.', 'checkpoint.pickle'),
                          help='checkpoint to FILE (default: ./checkpoint.pickle)')
    parser.add_argument('--verbose', '-v', dest='verbosity', action='count', default=0,
                        help='set verbose output; raise verbosity level with -vv, -vvv, -vvvv')
    parser.add_argument('--logging_config', '-l', type=str, default='logging.yaml',
                        help='set config file for logging; overides --verbose (default: logging.yaml)')
    AlgorithmFactory.add_options(parser.add_argument_group('algorithm'))
    group_breeding = parser.add_argument_group('breeding')
    MateFactory.add_options(group_breeding)
    MutateFactory.add_options(group_breeding)
    SelectFactory.add_options(group_breeding)
    CreateFactory.add_options(group_breeding)
    ParallelizationFactory.add_options(parser.add_argument_group('parallel execution'))

    args = parser.parse_args()

    workdir = os.path.dirname(os.path.abspath(args.checkpoint_file))
    if not os.path.exists(workdir):
        raise RuntimeError('Path does not exist: "{}"'.format(workdir))
    log_level = utils.logging.log_level(args.verbosity)
    utils.logging.load_config(config_file=args.logging_config, default_level=log_level,
                              placeholders=dict(workdir=workdir))
    logger = logging.getLogger(__name__)

    if args.resume_file is not None:
        logger.debug('Loading checkpoint {}'.format(args.resume_file))
        app = Application.from_checkpoint(args.resume_file)
        return app, args
    else:
        mate = MateFactory.create(args, IndividualClass)
        mutate = MutateFactory.create(args, IndividualClass)
        select = SelectFactory.create(args)
        create_method = CreateFactory.create(args, IndividualClass)
        algorithm_factory = toolz.partial(AlgorithmFactory.create, args, mate, mutate, select, create_method)
        parallel_factory = toolz.partial(ParallelizationFactory.create, args)
        assessment_runner = AssessmentRunnerClass(parallel_factory)
        gp_runner = GPRunner(IndividualClass, algorithm_factory, assessment_runner)
        app = Application(args, gp_runner, args.checkpoint_file, callbacks=callbacks)
        return app, args


class AFactory(object):
    _mapping = {}

    @classmethod
    def create(cls, config, *args, **kwargs):
        config = to_argparse_namespace(config)
        return cls._create(config, *args, **kwargs)

    @staticmethod
    def add_options(parser):
        """Add available parser options."""
        raise NotImplementedError

    @classmethod
    def get_from_mapping(cls, key):
        try:
            func = cls._mapping[key]
        except KeyError:
            raise RuntimeError('Option {} not supported'.format(key))
        return func


def get_mapping(group):
    return {obj.__name__.lower(): obj for obj in group}


class AlgorithmFactory(AFactory):
    """Factory class for gp algorithms."""

    _mapping = get_mapping(gp.all_algorithms)

    @classmethod
    def _create(cls, args, mate_func, mutate_func, select, create_func):
        """Setup gp algorithm."""
        kwargs = locals().copy()
        args.algorithm = args.algorithm.lower()
        algorithm_class = cls.get_from_mapping(args.algorithm)
        signature = inspect.signature(algorithm_class)
        algorithm = algorithm_class(*tuple(kwargs.get(x) for x in signature.parameters))
        algorithm.crossover_prob = args.crossover_prob
        algorithm.mutation_prob = args.mutation_prob
        return algorithm

    @staticmethod
    def add_options(parser):
        """Add available parser options."""
        parser.add_argument('--algorithm', dest='algorithm', type=str, default='nsga2',
                            choices=list(AlgorithmFactory._mapping.keys()),
                            help='the gp algorithm (default: nsga2)')
        parser.add_argument('--cxpb', dest='crossover_prob', metavar='p', type=utils.argparse.unit_interval,
                            default=0.5,
                            help='crossover probability for mating (default: 0.5)')
        parser.add_argument('--mutpb', dest='mutation_prob', metavar='p', type=utils.argparse.unit_interval,
                            default=0.2,
                            help='mutation probability (default: 0.2)')
        parser.add_argument('--tournament-size', dest='tournament_size', metavar='n', type=utils.argparse.unit_interval,
                            default=2,
                            help='tournament size for tournament selection (default: 2)')


class MateFactory(AFactory):
    """Factory class for gp mating functions."""

    _mapping = get_mapping(gp.all_crossover)

    @staticmethod
    def _create(args, IndividualClass):
        """Setup mating function."""
        args.mating = args.mating.lower()
        mate = MateFactory.get_from_mapping(args.mating)(**vars(args))
        static_limit_decorator = deap.gp.staticLimit(key=operator.attrgetter("height"),
                                                     max_value=args.mating_max_height)
        mate = static_limit_decorator(mate)
        return mate

    @staticmethod
    def add_options(parser):
        """Add available parser options."""
        parser.add_argument('--mating', dest='mating', type=str, default='cxonepoint',
                            choices=list(MateFactory._mapping.keys()),
                            help='the mating method (default: cxonepoint)')
        parser.add_argument('--mating-max-height', dest='mating_max_height', metavar='n',
                            type=utils.argparse.positive_int, default=20,
                            help='limit for the expression tree height as a result of mating (default: 20)')


class MutateFactory(AFactory):
    """Factory class for gp mutation functions."""

    _mapping = get_mapping(gp.all_mutations)

    @staticmethod
    def _create(args, IndividualClass):
        """Setup mutation function."""
        args.mutation = args.mutation.lower()
        mutate = MutateFactory.get_from_mapping(args.mutation)(IndividualClass.pset, **vars(args))
        static_limit_decorator = deap.gp.staticLimit(key=operator.attrgetter("height"),
                                                     max_value=args.mutation_max_height)
        mutate = static_limit_decorator(mutate)
        return mutate

    @staticmethod
    def add_options(parser):
        """Add available parser options."""
        parser.add_argument('--mutation', dest='mutation', type=str, default='mutuniform',
                            choices=list(MutateFactory._mapping.keys()),
                            help='the mutation method (default: mutuniform)')
        parser.add_argument('--mutation-max-height', dest='mutation_max_height', metavar='n',
                            type=utils.argparse.positive_int, default=20,
                            help='limit for the expression tree height as a result of mutation (default: 20)')
        parser.add_argument('--mutate_tree_min', dest='mutate_tree_min', default=0, metavar='min_',
                            type=utils.argparse.positive_int,
                            help="minimum value for tree based mutation methods (default: 0)")
        parser.add_argument('--mutate_tree_max', dest='mutate_tree_max', default=2, metavar='max_',
                            type=utils.argparse.positive_int,
                            help="maximum value for tree based mutation methods (default: 2)")


class SelectFactory(AFactory):
    """Factory class for selection"""

    _mapping = {"nsga2": deap.tools.selNSGA2,
                "spea2": deap.tools.selSPEA2,
                }

    @staticmethod
    def _create(args):
        args.select = args.select.lower()
        return SelectFactory.get_from_mapping(args.select)

    @staticmethod
    def add_options(parser):
        """Add available parser options."""
        parser.add_argument('--select', dest='select', type=str, default='nsga2',
                            choices=list(SelectFactory._mapping.keys()),
                            help='the selection method (default: nsga2)')


class CreateFactory(AFactory):
    """Factory class for creation"""
    _mapping = {"halfandhalf": deap.gp.genHalfAndHalf}

    @staticmethod
    def _create(args, IndividualClass):
        args.create_method = args.create_method.lower()
        m = CreateFactory.get_from_mapping(args.create_method)
        create_ = toolz.partial(IndividualClass.create_population, gen_method=m,
                                min=args.create_min_height, max=args.create_max_height)
        return create_

    def add_options(parser):
        parser.add_argument('--create_method', dest='create_method', type=str, default='halfandhalf',
                            choices=list(CreateFactory._mapping.keys()),
                            help='the create method (default: halfandhalf)')
        parser.add_argument('--create_max_height', dest='create_max_height', default=4,
                            type=utils.argparse.positive_int,
                            help="maximum value for tree based create methods (default: 4)")
        parser.add_argument('--create_min_height', dest='create_min_height', default=1,
                            type=utils.argparse.positive_int,
                            help="maximum value for tree based create methods (default: 1)")


class ParallelizationFactory(AFactory):
    """Factory class for parallel execution schemes."""

    @staticmethod
    def _create(args):
        return SingleProcessFactoy

    @staticmethod
    def add_options(parser):
        """Add available parser options."""
        # todo


def safe(file_name, **kwargs):
    """Dump kwargs to file."""
    with open(file_name, 'wb') as file:
        dill.dump(kwargs, file)


def load(file_name):
    """Load data saved with safe()."""
    with open(file_name, 'rb') as file:
        cp = dill.load(file)
    return cp


def create_tmp_dir(prefix='run-'):
    """Create directory with current time as signature."""
    start_date_str = time.strftime('%Y-%m-%d-%H%M%S', time.localtime())
    workdir = os.path.relpath(prefix + start_date_str)
    os.mkdir(workdir)
    return workdir


def _create_logger(verbosity, config_file, workdir):
    log_level = utils.logging.log_level(verbosity)
    utils.logging.load_config(config_file=config_file, default_level=log_level, placeholders=dict(workdir=workdir))
    return logging.getLogger(__name__)


def create_stats(n):
    """Create deap.tools.MultiStatistics object for n fitness values."""

    def val(i, ind):
        return ind.fitness.values[i]

    stats = dict()
    for i in range(n):
        stats['fit{}'.format(i)] = deap.tools.Statistics(toolz.partial(val, i))
    mstats = deap.tools.MultiStatistics(**stats)
    mstats.register("min", np.nanmin)
    mstats.register("max", np.nanmax)
    return mstats


def to_argparse_namespace(d):
    """Return argparse.Namespace object created from dictionary d."""
    if isinstance(d, argparse.Namespace):
        return d
    elif isinstance(d, dict):
        return argparse.Namespace(**d)
    else:
        raise RuntimeError('Cannot convert {} to argparse.Namespace.'.format(type(d)))



# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL

from .individual import simplify_this, AExpressionTree


class NullSpace:    # todo documentation
    def __init__(self, zero=True, infty=True, constant=False):
        self.zero = zero
        self.infty = infty
        self.constant = constant

    def __contains__(self, element):
        expr = simplify_this(element)
        if self.infty:
            if "zoo" in str(expr):
                return True

        if self.zero:
            if expr.is_zero:
                return  True

        if self.constant:
            if expr.is_constant():
                return True

            elif all(t.name in element.pset.constants for t in element.terminals):
                return True
        return False


def build_constraints(null_space, n_trials=30):
    """Create constraints decorators based on rules.

    :param null_space:
    :param n_trials: Number of tries. Give up afterwards (return input).

    :return: list of constraint decorators
    """
    def reject(operator):
        def inner(*inds, **kw):
            for i in range(n_trials):
                out = operator(*inds, **kw)
                if isinstance(out, AExpressionTree):  # can this be done w/o type checking?
                    t = out
                elif isinstance(out, (list, tuple)):
                    t = out[0]
                else:
                    raise RuntimeError
                if not t in null_space:
                    break
            else:
                if inds:
                    return inds
                else:
                    raise UserWarning
            return out
        return inner
    return [reject]


def apply_constraints(funcs, constraints):
    """Decorate a list of genetic operators with constraints.

    :param funcs: list of operators (mate, mutate, create)
    :param constraints: list of constraint decorators
    :return: constrained operators
    """
    for c in constraints:
        swap = [c(func) for func in funcs]
        funcs = swap
    return funcs



# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL

import time
import signal
import functools

import numpy as np


class SoftTimeOut:
    """Break condition based on a soft time out.
    Start a new generation as long as there is some time left.
    """
    def __init__(self, ttl):
        """

        :param ttl: time to live in seconds
        """
        self.soft_ttl = ttl
        self.start = time.time()

    @property
    def now(self):
        return time.time() - self.start

    @property
    def alive(self):
        return bool(self.soft_ttl - self.now > 0) or self.soft_ttl == 0

    def __call__(self, *args, **kwargs):
        return self.alive


def timeout(ttl):
    """ Decorate a function. Will raise `TimeourError` if function call takes longer than the ttl.

    Vendored from ffx.

    :param ttl: time to live in seconds
    """
    def decorate(f):
        def handler(signum, frame):
            raise TimeoutError()

        @functools.wraps(f)
        def new_f(*args, **kwargs):
            old = signal.signal(signal.SIGALRM, handler)
            signal.alarm(ttl)
            try:
                result = f(*args, **kwargs)
            finally:
                signal.signal(signal.SIGALRM, old)
            signal.alarm(0)
            return result
        return new_f
    return decorate


def max_fitness_on_timeout(max_fitness):
    """ Decorate a function. Associate max_fitness with long running individuals.

    :param max_fitness: fitness of aborted individual calls.
    :returns: fitness or max_fitness
    """
    def decorate(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            try:
                fitness = f(*args, **kwargs)
            except TimeoutError:
                fitness = max_fitness
            return fitness
        return inner
    return decorate


def soft_max_iter(app, max_iter=np.infty):
    """
    Soft breaking condition. Will check after each generation weather maximum number of iterations is exceeded.

    :type app: `glyph.application.Application`
    :param max_iter: maximum number of function evaluations
    :return: bool(iter) > max_iter
    """
    return sum(app.gp_runner.logbook.select("evals")) >= max_iter


def soft_target(app, target=0, error_index=0):
    """Soft breaking condition. Will check after each generation minimum error is reached.

    :type app: `glyph.application.Application`
    :param target: value of desired error metric
    :param error_index: index in fitness tuple
    :return: bool(min_error) <= target
    """
    return app.gp_runner.logbook.chapters["fit{}".format(error_index)].select("min")[-1] <= target


def break_condition(target=0, error_index=0, ttl=0, max_iter=np.infty):
    """Combined breaking condition based on time to live, minimum target and maximum number of iterations.

     :Example:
     >>> app = glyph.application.Application(...)
     >>> bc = break_condition(...)
     >>> app.run(break_condition=bc)


    :param target: value of desired error metric
    :param error_index: index in fitness tuple
    :param ttl: time to live in seconds
    :param max_iter: maximum number of iterations
    """

    sto = SoftTimeOut(ttl)
    def cb(app):
        return soft_max_iter(app, max_iter=max_iter) or sto(app) or soft_target(app, target=target, error_index=error_index)
    return cb



# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL

import toolz
import deap.gp
import itertools
import numpy as np


def mutuniform(pset, **kwargs):
    """Factory for mutuniform
    """
    min_ = kwargs.get('min_', 0)
    max_ = kwargs.get('max_', 2)
    expr_mut = toolz.partial(deap.gp.genFull, min_=min_, max_=max_)
    mutate = toolz.partial(deap.gp.mutUniform, expr=expr_mut, pset=pset)
    return mutate


def mutnodereplacement(pset, **kwargs):
    """Factory for mutnodereplacement"""
    return toolz.partial(deap.gp.mutNodeReplacement, pset=pset)


def mutinsert(pset, **kwargs):
    """Factory for mutinsert"""
    return toolz.partial(deap.gp.mutInsert, pset=pset)


def mutshrink(pset, **kwargs):
    """Factory for mutshrink"""
    return deap.gp.mutShrink


def cxonepoint(**kwargs):
    """Factory for cxonepoint"""
    return deap.gp.cxOnePoint


def cxonepointleafbiased(**kwargs):
    """Factory for cxonepointleafbiased"""
    termpb = kwargs.get('termpb', 0.1)
    return toolz.partial(deap.gp.cxOnePointLeafBiased, termpb=termpb)


def nd_mutation(atree, mut1d, rng=np.random):
    """

    :param atree:
    :type atree: `glyph.gp.individual.ANDimTree`
    :param mut1d: any mutation operator worklng for `glyph.gp.individual.AExpressionTree`
    :param rng: (seeded) random number generator
    :return: mutated offspring
    """
    a = rng.randint(0, atree.dim)
    atree[a] = mut1d(atree[a])[0]
    return atree,


def nd_crossover(atree, btree, cx1d, rng=np.random):
    """

    :param atree:
    :type atree: `glyph.gp.individual.ANDimTree`
    :param btree:
    :type btree: `glyph.gp.individual.ANDimTree`
    :param cx1d: any crossover operator working for `glyph.gp.individual.AExpressionTree`
    :param rng: (seeded) random number generator
    :return: two mated offsprings
    """
    n = atree.dim
    a, b = rng.randint(0, n, size=2)

    atree[a], btree[b] = cx1d(atree[a], btree[b])

    return atree, btree


all_mutations = [mutuniform, mutnodereplacement, mutshrink, mutshrink]
all_crossover = [cxonepoint, cxonepointleafbiased]
__all__ = [obj.__name__ for obj in itertools.chain(all_mutations, all_crossover)]



# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL


"""Provide Individual class for gp."""

import re
import copy
import deap.gp
import deap.base
import sympy
import functools
import itertools
import numpy as np

import glyph.utils

#def len_subtree(i):
#    sl_left = ind.searchSubtree(i+1)
#    len_left = sl_left.stop - sl_left.start
#    sl_right = ind.searchSubtree(sl_left.stop)
#    len_right = sl_right.stop - sl_right.start
#    return len_left, len_right


def sc_qout(x, y):
    """SC is the quotient of the number of nodes of its left and right child-trees x and y"""
    return x / y


def sc_mmqout(x, y, cmin=-1, cmax=1):
    """SC is the minimum-maximum quotient of the number of nodes of both
    child-trees x and y mapped into the constant interval [cmin, cmax]"""
    return cmin + min(x, y)/max(x, y) * (cmax - cmin)


class StructConst(deap.gp.Primitive):
    def __init__(self, func):
        """
        :param func: evaluate left and right subtree and assign a constant.
        """
        self.func = func
        super().__init__("SC", [deap.gp.__type__]*2, deap.gp.__type__)

    @staticmethod
    def get_len(expr, tokens=("(,")):
        regex = "|".join("\\{}".format(t) for t in tokens)
        return len(re.split(regex, expr))

    def format(self, *args):
        left, right = args
        return str(self.func(self.get_len(left), self.get_len(right)))


def add_sc(pset, func):
    """Adds a structural constant to a given primitive set.
    :param func: `callable(x, y) -> float` where x and y are the expressions of the left and right subtree
    :param pset: You may want to use `sympy_primitive_set` or `numpy_primitive_set` without symbolic constants.
    :type pset: `deap.gp.PrimitiveSet`
    """
    sc = StructConst(func)
    pset._add(sc)
    pset.prims_count += 1
    return pset


def _build_args_string(pset, consts):
    args = ','.join(arg for arg in pset.args)
    if consts:
        if args:
            args += ','
        args += ','.join("{}=1.0".format(arg) for arg in consts)
    return args


def sympy_primitive_set(categories=('algebraic', 'trigonometric', 'exponential'), arguments=['y_0'], constants=[]):
    """Create a primitive set with sympy primitves.

    :param arguments: variables to use in primitive set
    :param constants: symbolic constants to use in primitive set
    :param categories: an optional list of function categories for the primitive set. The following are available
                       'algebraic', 'neg', 'trigonometric', 'exponential', 'exponential', 'logarithm', 'sqrt'.

     :return: `deap.gp.PrimitiveSet`
    """
    pset = deap.gp.PrimitiveSet('main', arity=0)
    if 'algebraic' in categories:
        pset.addPrimitive(sympy.Add, arity=2)
        pset.addPrimitive(sympy.Mul, arity=2)
    if 'neg' in categories:
        pset.addPrimitive(functools.partial(sympy.Mul, -1.0), arity=1, name='Neg')
    if 'trigonometric' in categories:
        pset.addPrimitive(sympy.sin, arity=1)
        pset.addPrimitive(sympy.cos, arity=1)
    if 'exponential' in categories:
        pset.addPrimitive(sympy.exp, arity=1)
    if 'logarithm' in categories:
        pset.addPrimitive(sympy.ln, arity=1)
    if 'sqrt' in categories:
        pset.addPrimitive(sympy.sqrt, arity=1)

    # Use sympy symbols for argument representation.
    for symbol in itertools.chain(arguments, constants):
        pset.addTerminal(sympy.Symbol(symbol), name=symbol)

    pset.args = arguments
    pset.constants = constants
    return pset


def sympy_phenotype(individual):
    """Compile python function from individual.

    Uses sympy's lambdify function. Terminals from the primitive set will be
    used as parameters to the constructed lambda function; primitives (like
    sympy.exp) will be converted into numpy expressions (eg. numpy.exp).

    :param individual:
    :type individual: `glyph.gp.individual.AExpressionTree`
    :return: lambda function
    """
    pset = individual.pset
    args = _build_args_string(pset, pset.constants)
    expr = sympy.sympify(deap.gp.compile(repr(individual), pset))
    func = sympy.utilities.lambdify(args, expr, modules='numpy')
    return func


def numpy_primitive_set(arity, categories=('algebraic', 'trigonometric', 'exponential', 'symc')):
    """Create a primitive set based on numpys vectorized functions.


    :param arity: Number of variables in the primitive set
    :param categories:
    :return: `deap.gp.PrimitiveSet`

    .. note :: All functions will be closed, that is non-defined values will be mapped to 1. 1/0 = 1!
    """
    pset = deap.gp.PrimitiveSet("main", arity)
    # Use primitive set built-in for argument representation.
    pset.renameArguments(**{'ARG{}'.format(i): 'x_{}'.format(i) for i in range(arity)})
    pset.args = pset.arguments
    if 'symc' in categories:
        symc = 1.0
        pset.addTerminal(symc, "Symc")
        pset.constants = ["Symc"]
    else:
        pset.constants = []

    def close_function(func, value):
        @functools.wraps(func)
        def closure(*args):
            res = func(*args)
            if isinstance(res, np.ndarray):
                res[np.isinf(res)] = value
                res[np.isnan(res)] = value
            elif np.isinf(res) or np.isnan(res):
                res = value
            return res
        return closure

    if 'algebraic' in categories:
        pset.addPrimitive(np.add, 2, name="Add")
        pset.addPrimitive(np.subtract, 2, name="Sub")
        pset.addPrimitive(np.multiply, 2, name="Mul")
        _div = close_function(np.divide, 1)
        pset.addPrimitive(_div, 2, name="Div")
    if 'trigonometric' in categories:
        pset.addPrimitive(np.sin, arity=1)
        pset.addPrimitive(np.cos, arity=1)
    if 'exponential' in categories:
        pset.addPrimitive(np.exp, arity=1)
        _log = close_function(np.log, 1)
        pset.addPrimitive(_log, arity=1)
    if 'sqrt' in categories:
        _sqrt = close_function(np.sqrt, 1)
        pset.addPrimitive(_sqrt, arity=1)
    if "constants" in categories:
        pset.addTerminal(np.pi, name="pi")
        pset.addTerminal(np.e, name="e")
    return pset


def _get_index(ind, c):
    return [i for i, node in enumerate(ind) if node.name == c]


def numpy_phenotype(individual):
    """ Lambdify the individual

    :param individual:
    :type individual: `glyph.gp.individual.AExpressionTree`
    :return: lambda function

    :Note:
    In constrast to sympy_phenotype the callable will have a variable number of keyword arguments depending on the number
    of symbolic constants in the individual.

    :Example:
    >>> pset = numpy_primitive_set(1)
    >>> class Individual(AExpressionTree)
    >>>    pset = pset
    >>> ind = Individual.from_string("Add(x_0, symc)")
    >>> f = numpy_phenotype(ind)
    >>> f(1, 1)
    2
    """
    pset = individual.pset
    if pset.constants:
        c = pset.constants[0]
        index = _get_index(individual, c)
    else:
        index = []
    consts = ["c_{}".format(i) for i in range(len(index))]
    args = _build_args_string(pset, consts)
    expr = repr(individual)
    for c_ in consts:
        expr = expr.replace(c, c_, 1)
    func = sympy.utilities.lambdify(args, expr, modules=pset.context)
    return func


class AExpressionTree(deap.gp.PrimitiveTree):
    """Abstract base class for the genotype.

    Derived classes need to specify a primitive set from which the expression
    tree can be build, as well as a phenotype method.
    """

    pset = None
    hasher = str

    def __init__(self, content):
        if self.pset is None:
            raise TypeError("Cannot instantiate abstract {} with abstract attribute pset.".format(self.__class__))
        super(AExpressionTree, self).__init__(content)
        self.fitness = Measure()

    def __repr__(self):
        """Symbolic representation of the expression tree."""
        repr = ''
        stack = []
        for node in self:
            stack.append((node, []))
            while len(stack[-1][1]) == stack[-1][0].arity:
                prim, args = stack.pop()
                repr = prim.format(*args)
                if len(stack) == 0:
                    break
                stack[-1][1].append(repr)
        return repr

    def __hash__(self):
        return hash(self.hasher(self))

    def __eq__(self, other):
        return hash(self) == hash(other)

    @property
    def const_opt(self):
        return getattr(self, "popt", ())

    @property
    def terminals(self):
        """Return terminals that occur in the expression tree."""
        return [primitive for primitive in self if primitive.arity == 0]

    @classmethod
    def from_string(cls, string):
        return super(AExpressionTree, cls).from_string(string, cls.pset)

    @classmethod
    def create(cls, gen_method=deap.gp.genHalfAndHalf, min=1, max=4):
        return cls(gen_method(cls.pset, min_=min, max_=max))

    @classmethod
    def create_population(cls, size, gen_method=deap.gp.genHalfAndHalf, min=1, max=4):
        """Create a list of individuals of class Individual."""
        if size < 0:
            raise RuntimeError('Cannot create population of size {}'.format(size))
        return [cls.create(gen_method=gen_method, min=min, max=max) for _ in range(size)]


def nd_phenotype(nd_tree, backend=sympy_phenotype):
    """
    :param nd_tree:
    :type  nd_tree: ANDimTree
    :param backend: sympy_phenotype or numpy_phenotype
    :return: lambda function
    """
    funcs = [backend(t) for t in nd_tree]
    return lambda *x: [f(*x) for f in funcs]


class ANDimTree(list):
    """
    A naive tree class representing a vector-valued expression. Each dimension is encoded as a expression tree.
    """
    def __init__(self, trees):
        super().__init__(trees)
        self.dim = len(trees)
        self.fitness = Measure()

    @property
    def base(self):
        pass

    def __repr__(self):
        return str([str(tree) for tree in self])

    @classmethod
    def create_individual(cls, ndim):
        return cls(cls.base.create_population(ndim))

    @classmethod
    def create_population(cls, size, ndim):
        return [cls.create_individual(ndim) for _ in range(size)]

    @property
    def height(self):
        return len(self)

    @property
    def pset(self):
        return self.base.pset

    @property
    def terminals(self):
        """Return terminals that occur in the expression tree."""
        return [primitive for primitive in itertools.chain.from_iterable(self) if primitive.arity == 0]

    @classmethod
    def from_string(cls, strs):
        return cls([cls.base.from_string(s) for s in strs])


class Measure(deap.base.Fitness):
    """
    This is basically a wrapper around deap.base.Fitness.

    It provides the following enhancements over the base class:
    - more adequate naming
    - copy constructable
    - no weight attribute
    """

    weights = itertools.repeat(-1)

    def __init__(self, values=()):
        if len(values) > 0:
            self.values = values

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def get_values(self):
        return tuple(-1.0 * v for v in self.wvalues)

    def set_values(self, values):
        self.wvalues = tuple(-1.0 * v for v in values)

    def del_values(self):
        self.wvalues = ()

    values = property(get_values, set_values, del_values)


def convert_inverse_prim(prim, args):
    """
    Convert inverse prims according to:
    [Dd]iv(a,b) -> Mul[a, 1/b]
    [Ss]ub(a,b) -> Add[a, -b]

    We achieve this by overwriting the corresponding format method of the sub and div prim.
    """
    prim = copy.copy(prim)
    prim.name = re.sub(r'([A-Z])', lambda pat: pat.group(1).lower(), prim.name)    # lower all capital letters

    converter = {
        'sub': lambda *args_: "Add({}, Mul(-1,{}))".format(*args_),
        'div': lambda *args_: "Mul({}, Pow({}, -1))".format(*args_)
    }
    prim_formatter = converter.get(prim.name, prim.format)

    return prim_formatter(*args)


def stringify_for_sympy(f):
    """Return the expression in a human readable string.
    """
    string = ""
    stack = []
    for node in f:
        stack.append((node, []))
        while len(stack[-1][1]) == stack[-1][0].arity:
            prim, args = stack.pop()
            string = convert_inverse_prim(prim, args)
            if len(stack) == 0:
                break  # If stack is empty, all nodes should have been seen
            stack[-1][1].append(string)
    return string


@glyph.utils.Memoize
def simplify_this(expr):
    """
    :param expr:
    :type expr: str
    :return: Sympy representation of simplified expression#

    :warning: does not respect closures
    """
    return sympy.simplify(stringify_for_sympy(expr))


def child_trees(ind):
    start = 1
    while start < len(ind):
        slice_ = ind.searchSubtree(start)
        yield type(ind)(ind[slice_])
        start += slice_.stop - 1


@glyph.utils.Memoize
def simplify_constant(ind):
    """Trims subtrees of symbolic constants.

    Recursively applying these rules:

        S f(*[symc]*f.arity) = symc
        S f(x,... symc) = f(x,... symc)
        S symc = symc

    For deeper trees, try to trim down lower levels first.
    An individual cannot be trimmed down further if its a fixpoint of S.
    """
    symc = ind.pset.mapping[ind.pset.constants[0]]
    if symc in ind:
        root = ind.root

        # the tree is just a function of constants
        if len(ind) > 1 and  all(i == symc for i in ind[1:]):
            return type(ind)([symc])

        # root is just a constant or tree is a function of a variable and cannot be trimmed down.
        elif len(ind) == 1 or len(ind) == root.arity + 1:
            return ind

        # try to simplify all children of root
        else:
            acc = [simplify_constant(child) for child in child_trees(ind)]
            new_ind = type(ind)([root] + sum(acc, []))

            # cannot be simplified
            if ind == new_ind:
                return ind

            # can it be simplified further?
            else:
                return simplify_constant(new_ind)

    else:
        return ind


def pretty_print(expr, constants, consts_values, count=0):
    """Replace symbolic constants in the str representation of an individual
    by their numeric values.

    This checks for either c followed by ")" or ",", or for just the constant itself.
    """
    for k, v in zip(constants, consts_values):
        pattern = r"{c}(?=[,)])|^{c}$".format(c=str(k))
        expr = re.sub(pattern, str(v), expr, count=count)
    return expr



# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL

from . import argparse
from . import logging
from . import numeric
from . import break_condition


class Memoize:
    """Memoize(fn) - an instance which acts like fn but memoizes its arguments
       Will only work on functions with non-mutable arguments
       http://code.activestate.com/recipes/52201/
    """
    def __init__(self, fn):
        self.fn = fn
        self.memo = {}

    def __call__(self, *args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in self.memo:
            self.memo[key] = self.fn(*args, **kwargs)
        return self.memo[key]



# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL

import itertools
import functools
from collections import defaultdict

import numpy as np
import scipy.integrate
import scipy.optimize


def integrate(dy, yinit, x, f_args=(), integrator='dopri5', **integrator_args):
    """
    Convenience function for odeint().

    Uselful if you do not want to step through the integration, but rather get
    the full result in one call.

    :param dy: `callable(x, y, *args)`
    :param yinit: sequence of initial values.
    :param x: sequence of x values.
    :param f_args: (optional) extra arguments to pass to function.
    :returns: y(x)
    """
    res = odeint(dy, yinit, x, f_args=f_args, integrator=integrator, **integrator_args)
    y = np.vstack(res).T
    if y.shape[1] != x.shape[0]:
        y = np.empty((y.shape[0], x.shape[0]))
        y[:] = np.NAN
    if y.shape[0] == 1:
        return y[0, :]
    return y


def odeint(dy, yinit, x, f_args=(), integrator='dopri5', **integrator_args):
    """
    Integrate the initial value problem (dy, yinit).

    A wrapper around scipy.integrate.ode.

    :param dy: `callable(x, y, *args)`
    :param yinit: sequence of initial values.
    :param x: sequence of x values.
    :param f_args: (optional) extra arguments to pass to function.
    :yields: y(x_i)
    """
    @functools.wraps(dy)
    def f(x, y):
        return dy(x, y, *f_args)
    ode = scipy.integrate.ode(f)
    ode.set_integrator(integrator, **integrator_args)
    # ode.set_f_params(*f_args)
    # Seems to be buggy! (see https://github.com/scipy/scipy/issues/1976#issuecomment-17026489)
    ode.set_initial_value(yinit, x[0])
    yield ode.y
    for value in itertools.islice(x, 1, None):
        ode.integrate(value)
        if not ode.successful():
            raise StopIteration
        yield ode.y


def rms(y):
    """Root mean square."""
    return np.sqrt(np.mean(np.square(y)))


def strict_subtract(x, y):
    try:
        if x.shape != y.shape:
            raise ValueError('operands could not be broadcast together with shapes {} {}'.format(x.shape, y.shape))
    except AttributeError:
        pass
    return x - y


def rmse(x, y):
    """Root mean square error."""
    return rms(strict_subtract(x, y))


def nrmse(x, y):
    """Normalized, with respect to x, root mean square error."""
    diff = strict_subtract(x, y)
    return rms(diff) / (np.max(x) - np.min(x))


def cvrmse(x, y):
    """Coefficient of variation, with respect to x, of the rmse."""
    diff = strict_subtract(x, y)
    return rms(diff) / np.mean(x)


def silent_numpy(func):
    @functools.wraps(func)
    def closure(*args, **kwargs):
        with np.errstate(all='ignore'):
            return func(*args, **kwargs)
    return closure


def hill_climb(fun, x0, args, precision=5, maxfev=100, directions=5, target=0, rng=np.random, **kwargs):
    """Stochastic hill climber for constant optimization.
    Try self.directions different solutions per iteration to select a new best individual.

    :param fun: function to optimize
    :param x0: initial guess
    :param args: additional arguments to pass to fun
    :param precision: maximum precision of x0
    :param maxfev: maximum number of function calls before stopping
    :param directions: number of directions to explore before doing a hill climb step
    :param target: stop if fun(x) <= target
    :param rng: (seeded) random number generator

    :return: `scipy.optimize.OptimizeResult`
    """
    res = scipy.optimize.OptimizeResult()

    def tweak(x):
        """ x = round(x + xi, p) with xi ~ N(0, sqrt(x)+10**(-p))"""
        return round(x+rng.normal(scale=np.sqrt(abs(x))+10**(-precision)), precision)

    def f(x):
        return fun(x, *args)

    x = x0
    fx = f(x)
    it = 1
    if len(x0) > 0:
        while fx >= target and it <= maxfev:
            memory = [(x, fx)]
            for j in range(directions):
                it += 1
                xtweak = np.array([tweak(c) for c in x])
                fxtweak = f(xtweak)
                memory.append((xtweak, fxtweak))
                if fxtweak <= target:
                    break
            x, fx = min(memory, key=lambda t: t[1])

    res["x"] = x
    res["fun"] = fx
    res["success"] = True
    res["nfev"] = it

    return res


class SmartConstantOptimizer:
    """Decorate a minimize method used in `scipy.optimize.minimize` to cancel non promising constant optimizations.

    The stopping criteria is based on the improvment rate :math:`\frac{\Delta f}[\Delta fev}`.

    If the improvment rate is below the :math:`q_{threshold}` quantile for a given number of function
    evaluations, optimization is stopped.
    """
    def __init__(self, method, step_size=10, min_stat=10, threshold=25):
        """
        :params method: see `scipy.optimize.minimize` method
        :params step_size: number of function evaluations betweem iterations
        :params min_stat: minmum sample size before stopping
        :params threshold: quantile
        """
        self.method = method
        self.step_size = step_size
        self.min_stat = min_stat
        self.threshold = threshold
        self.memory = defaultdict(list)

    def __call__(self, fun, x0, args, **kwargs):

        maxfev = kwargs.get('maxfev', 1000*len(x0))
        kw = kwargs.copy()
        kw["maxfev"] = self.step_size

        res = self.method(fun, x0, args, **kw)
        fx_base = res.fun
        x0 = res.x

        fev = self.step_size
        while fev <= maxfev - self.step_size:

            res = self.method(fun, x0, args, **kw)
            fev += res.nfev
            fx = res.fun
            x0 = res.x

            eps = (fx - fx_base) / fev
            self.memory[fev].append(eps)
            if len(self.memory[fev]) > self.min_stat:
                if eps < np.percentile(self.memory[fev], self.threshold):
                    break

        return res


def expressional_complexity(ind):
    """Sum of length of all subtrees of the individual."""
    return sum(len(ind[ind.searchSubtree(i)]) for i in range(len(ind)))



# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL

import os
import json
import logging
import random
import argparse
import copy
import itertools
from functools import partial, wraps
from threading import Thread
import concurrent.futures
from queue import Queue
from time import sleep

import zmq
import yaml
import deap.tools
import deap.gp
import toolz
import numpy as np
from scipy.optimize._minimize import _minimize_neldermead as nelder_mead
from cache import DBCache

from glyph.gp import AExpressionTree
from glyph.utils.logging import print_params
from glyph.utils.argparse import readable_file
from glyph.utils.break_condition import break_condition
from glyph.assessment import tuple_wrap, const_opt_scalar
from glyph.gp.individual import simplify_this, add_sc, sc_mmqout, pretty_print
from glyph.gp.constraints import build_constraints, apply_constraints, NullSpace
import glyph.application
import glyph.utils


def partition(pred, iterable):
    """Use a predicate to partition entries into false entries and true entries"""
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    t1, t2 = itertools.tee(iterable)
    return itertools.filterfalse(pred, t1), filter(pred, t2)


class RemoteApp(glyph.application.Application):
    def run(self, break_condition=None):
        """For details see application.Application.
        Will checkpoint and close zmq connection on keyboard interruption.
        """
        try:
            super().run(break_condition=break_condition)
        except KeyboardInterrupt:
            self.checkpoint()
        finally:
            self.assessment_runner.send(dict(action="SHUTDOWN"))
            zmq.Context.instance().destroy()

    @classmethod
    def from_checkpoint(cls, file_name, send, recv):
        """Create application from checkpoint file."""
        cp = glyph.application.load(file_name)
        gp_runner = cp['runner']
        gp_runner.assessment_runner = RemoteAssessmentRunner(send, recv, consider_complexity=cp['args'].consider_complexity,
                                                            method=cp['args'].const_opt_method, options=cp['args'].options,
                                                            caching=cp['args'].caching, simplify=cp['args'].simplify,
                                                            persistent_caching=cp['args'].persistent_caching, chunk_size=cp['args'].chunk_size)
        app = cls(cp['args'], cp['runner'], file_name, cp['callbacks'])
        app.pareto_fronts = cp['pareto_fronts']
        app._initialized = True
        random.setstate(cp['random_state'])
        return app

    def checkpoint(self):
        """Checkpoint current state of evolution."""

        runner = copy.deepcopy(self.gp_runner)
        del runner.assessment_runner
        glyph.application.safe(self.checkpoint_file, args=self.args, runner=runner,
                               random_state=random.getstate(), pareto_fronts=self.pareto_fronts, callbacks=self.callbacks)
        self.logger.debug('Saved checkpoint to {}'.format(self.checkpoint_file))


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', type=int, default=5555, help='Port for the zeromq communication (default: 5555)')
    parser.add_argument('--ip', type=str, default="localhost", help='IP of the client (default: localhost)')

    config = parser.add_argument_group('config')
    group = config.add_mutually_exclusive_group()
    group.add_argument('--remote', action='store_true', dest='remote', default=False, help='Request GP configs from experiment handler.')
    group.add_argument('--cfile', dest='cfile', type=readable_file, help='Read GP configs from file')

    RemoteApp.add_options(parser)
    cp_group = parser.add_mutually_exclusive_group(required=False)
    cp_group.add_argument('--ndim', type=int, default=1)
    cp_group.add_argument('--resume', dest='resume_file', metavar='FILE', type=str,
                          help='continue previous run from a checkpoint file')
    cp_group.add_argument('-o', dest='checkpoint_file', metavar='FILE', type=str,
                          default=os.path.join('.', 'checkpoint.pickle'),
                          help='checkpoint to FILE (default: ./checkpoint.pickle)')
    parser.add_argument('--verbose', '-v', dest='verbosity', action='count', default=0,
                        help='set verbose output; raise verbosity level with -vv, -vvv, -vvvv')
    parser.add_argument('--logging', '-l', dest='logging_config', type=str, default='logging.yaml',
                        help='set config file for logging; overides --verbose (default: logging.yaml)')
    glyph.application.AlgorithmFactory.add_options(parser.add_argument_group('algorithm'))
    group_breeding = parser.add_argument_group('breeding')
    glyph.application.MateFactory.add_options(group_breeding)
    glyph.application.MutateFactory.add_options(group_breeding)
    glyph.application.SelectFactory.add_options(group_breeding)
    glyph.application.CreateFactory.add_options(group_breeding)

    ass_group = parser.add_argument_group('assessment')
    #ass_group.add_argument('--send_all', action='store_true', default=False, help='Send all invalid individuals at once. (default: False)')
    ass_group.add_argument('--simplify', type=bool, default=True, help='Simplify expression before sending them. (default: True)')
    ass_group.add_argument('--consider_complexity', type=bool, default=True, help='Consider the complexity of solutions for MOO (default: True)')
    ass_group.add_argument('--caching', type=bool, default=True, help='Cache evaluation (default: True)')
    ass_group.add_argument('--persistent_caching', default=None, help='Key for persistent data base cache for caching between experiments (default: None)')
    ass_group.add_argument('--max_fev_const_opt', type=int, default=100, help='Maximum number of function evaluations for constant optimization (default: 100)')
    ass_group.add_argument('--directions', type=int, default=5, help='Directions for the stochastic hill-climber (default: 5 only used in conjunction with --const_opt_method hill_climb)')
    ass_group.add_argument('--precision', type=int, default=3, help='Precision of constants (default: 3)')
    ass_group.add_argument('--const_opt_method', choices=['hill_climb', 'Nelder-Mead'], default='Nelder-Mead', help='Algorithm to optimize constants given a structure (default: Nelder-Mead)')
    ass_group.add_argument('--structural_constants', action='store_true', default=False, help='Make use of structural constants. (default: False)')
    ass_group.add_argument('--sc_min', type=float, default=-1, help='Minimum value of sc for scaling. (default: -1)')
    ass_group.add_argument('--sc_max', type=float, default=1, help='Maximum value of sc for scaling. (default: 1)')
    ass_group.add_argument('--smart', action="store_true", default=False, help='Use smart constant optimization. (default: False)')
    ass_group.add_argument('--smart_step_size', type=int, default=10, help='Number of fev in iterative function optimization. (default: 10)')
    ass_group.add_argument('--smart_min_stat', type=int, default=10, help='Number of samples required prior to stopping (default: 10)')
    ass_group.add_argument('--smart_threshold', type=int, default=25, help='Quantile of improvement rate. Abort constant optimization if below (default: 25)')
    ass_group.add_argument('--chunk_size', type=int, default=30, help='Number of individuals send per single request. (default: 30)')


    break_condition = parser.add_argument_group('break condition')
    break_condition.add_argument('--ttl', type=int, default=-1, help='Time to life (in seconds) until soft shutdown. -1 = no ttl (default: -1)')
    break_condition.add_argument('--target', type=float, default=0, help='Target error used in stopping criteria (default: 0)')
    break_condition.add_argument('--max_iter_total', type=int, default=np.infty, help='Maximum number of function evaluations (default: np.infty)')

    constraints = parser.add_argument_group('constraints')
    constraints.add_argument('--constraints_zero', type=bool, default=True, help='Discard zero individuals (default: True)')
    constraints.add_argument('--constraints_constant', type=bool, default=True, help='Discard constant individuals (default: True)')
    constraints.add_argument('--constraints_infty', type=bool, default=True, help='Discard individuals with infinities (default: True)')
    return parser


def _send(socket, msg, serializer=json):
    socket.send(serializer.dumps(msg).encode('ascii'))


def _recv(socket, serializer=json):
    return serializer.loads(socket.recv().decode('ascii'))


def connect(ip, port):
    socket = zmq.Context.instance().socket(zmq.REQ)
    socket.connect('tcp://{ip}:{port}'.format(ip=ip, port=port))
    send = partial(_send, socket)
    recv = partial(_recv, socket)
    return send, recv


def handle_const_opt_config(args):
    smart_options = {"use": args.smart, "kw": {"threshold": args.smart_threshold, "step_size": args.smart_step_size, "min_stat": args.smart_min_stat}}
    options = {'maxfev': args.max_fev_const_opt, 'smart_options': smart_options}
    if args.const_opt_method == 'hill_climb':
        options['directions'] = args.directions
        options['precision'] = args.precision
        options['target'] = args.target
    else:
        options['xatol'] = 10**(-args.precision)
        options['fatol'] = args.target
    args.options = options
    return args


def update_namespace(ns, up):
    """Update the argparse.Namespace ns with a dictionairy up."""
    return argparse.Namespace(**{**vars(ns), **up})


def handle_gpconfig(config, send, recv):
    """Will try to load config from file or from remote and update the cli/default config accordingly."""
    if config.cfile:
        with open(config.cfile, 'r') as cf:
            gpconfig = yaml.load(cf)
    elif config.remote:
        send(dict(action="CONFIG"))
        gpconfig = recv()
    else:
        gpconfig = {}
    return update_namespace(config, gpconfig)


def build_pset_gp(primitives, structural_constants=False, cmin=-1, cmax=1):
    """Build a primitive set used in remote evaluation. Locally, all primitives correspond to the id() function.
    """
    pset = deap.gp.PrimitiveSet('main', arity=0)
    pset.constants = set()
    for fname, arity in primitives.items():
        if arity > 0:
            func = lambda *args: args
            pset.addPrimitive(func, arity, name=fname)
        elif arity == 0:
            pset.addTerminal(fname, name=fname)
            pset.arguments.append(fname)
        else:
            pset.addTerminal(fname, name=fname)
            pset.constants.add(fname)
    if len(pset.terminals) == 0:
        raise RuntimeError("Pset needs at least one terminal node. You may have forgotten to specify it.")

    if structural_constants:
        f = partial(sc_mmqout, cmin=cmin, cmax=cmax)
        pset = add_sc(pset, f)
    return pset


class MyQueue(Queue):
    def __init__(self, send, recv, result_queue, expect):
        self.recv = recv
        self.send = send
        self.result_queue = result_queue
        self.expect = expect
        self.logger = logging.getLogger(self.__class__.__name__)

        super().__init__()

    def run(self, chunk_size=100):
        payloads = []
        keys = []

        def process(keys, payloads):
            self.send(dict(action="EXPERIMENT", payload=payloads))
            fitnesses = self.recv()["fitness"]
            for key,fit in zip(keys, fitnesses):
                self.logger.debug("Writing result for key: {}".format(key))
                self.result_queue[key] = fit

        while self.expect > 0:
            key_payload = self.get()

            if key_payload is None:
                self.expect -= 1
                if self.expect == 0:
                    break
            else:
                key, payload = key_payload
                if key not in self.result_queue:
                    self.logger.debug("Queueing key: {}".format(key))
                    payloads.append(payload)
                    keys.append(key)
            if len(payloads) == min(self.expect, chunk_size):
                process(keys, payloads)
                payloads = []
                keys = []
        if payloads:
            process(keys, payloads)


def key_set(itr, key=hash):
    keys = map(key, itr)
    s = {k: v for k, v in zip(keys, itr)}
    return list(s.values())


class RemoteAssessmentRunner:
    def __init__(self, send, recv, consider_complexity=True, method='Nelder-Mead', options={'smart_options': {'use': False}}, caching=True, persistent_caching=None, simplify=True, chunk_size=30):
        """Contains assessment logic. Uses zmq connection to request evaluation."""
        self.send = send
        self.recv = recv
        self.consider_complexity = consider_complexity
        self.options = options
        self.caching = caching
        self.cache = {} if persistent_caching is None else DBCache("glyph-remote", persistent_caching)
        self.make_str = (lambda i: str(simplify_this(i))) if simplify else str
        self.result_queue = {}
        self.chunk_size = min(chunk_size, 30)
        self.method = {'hill_climb': glyph.utils.numeric.hill_climb}.get(method, nelder_mead)
        self.evaluations = 0

        self.smart_options = options.get('smart_options')
        if self.smart_options["use"]:
            self.method = glyph.utils.numeric.SmartConstantOptimizer(glyph.utils.numeric.hill_climb, **self.smart_options["kw"])

    def predicate(self, ind):
        """Does this individual need to be evaluated?"""
        return self.caching and self._hash(ind) in self.cache

    def _hash(self, ind):
        return json.dumps([self.make_str(t) for t in ind])

    def evaluate_single(self, individual, *consts):

        """Evaluate a single individual."""
        payload = [self.make_str(t) for t in individual]

        payload = [pretty_print(s, individual.pset.constants, consts) for s in payload]

        key = sum(map(hash, payload))   # constants may have been simplified, not in payload anymore.
        self.queue.put((key, payload))
        self.evaluations += 1

        result = None
        while result is None:
            sleep(0.1)
            result = self.result_queue.get(key)
        return result

    def measure(self, individual):
        """Construct fitness for given individual."""
        popt, error = const_opt_scalar(self.evaluate_single, individual, method=self.method, options=self.options)
        self.queue.put(None)
        individual.popt = popt
        if self.consider_complexity:
            fitness = error, sum(map(len, individual))
        else:
            fitness = error,
        return fitness

    def update_fitness(self, population):
        self.evaluations = 0

        invalid = [p for p in population if not p.fitness.valid]

        calculate, cached = map(list, partition(self.predicate, invalid))

        cached_fitness = [self.cache[self._hash(ind)] for ind in cached]
        calculate_duplicate_free = key_set(calculate, key=self._hash)
        # if we have duplicates in the calculate list, dont calculate these more than once.
        dup_free_cache = {}
        n = len(calculate_duplicate_free)
        if n > 0:             # main work is done here
            n_workers = min(n, self.chunk_size)

            # start queue and the broker
            self.queue = MyQueue(self.send, self.recv, self.result_queue, n)
            thread = Thread(target=self.queue.run, args=(n_workers,))
            thread.start()

            with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as exe:
                for k, future in zip(calculate_duplicate_free, exe.map(self.measure, calculate_duplicate_free)):
                    dup_free_cache[self._hash(k)] = future
            thread.join()
            del self.queue

            calculate_fitness = [dup_free_cache[self._hash(k)] for k in calculate]
        else:
            calculate_fitness = []

        # save to cache
        for key, fit in zip(map(self._hash, calculate), calculate_fitness):
            self.cache[key] = fit

        # assign fitness to individuals
        for ind, fit in zip(cached + calculate, cached_fitness + calculate_fitness):
            ind.fitness.values = fit

        return self.evaluations


    def __call__(self, population):
        return self.update_fitness(population)


class Individual(AExpressionTree):
    pass


class NDTree(glyph.gp.individual.ANDimTree):
    base = Individual

    def __hash__(self):
        return hash(hash(x) for x in self)


def make_remote_app(callbacks=()):
    parser = get_parser()
    args = parser.parse_args()

    send, recv = connect(args.ip, args.port)
    workdir = os.path.dirname(os.path.abspath(args.checkpoint_file))
    if not os.path.exists(workdir):
        raise RuntimeError('Path does not exist: "{}"'.format(workdir))
    log_level = glyph.utils.logging.log_level(args.verbosity)
    glyph.utils.logging.load_config(config_file=args.logging_config, default_level=log_level, placeholders=dict(workdir=workdir))
    logger = logging.getLogger(__name__)

    if args.resume_file is not None:
        logger.debug('Loading checkpoint {}'.format(args.resume_file))
        app = RemoteApp.from_checkpoint(args.resume_file, send, recv)
    else:
        args = handle_const_opt_config(handle_gpconfig(args, send, recv))
        try:
            pset = build_pset_gp(args.primitives, args.structural_constants, args.sc_min, args.sc_max)
        except AttributeError:
            raise AttributeError("You need to specify the pset")
        Individual.pset = pset
        mate = glyph.application.MateFactory.create(args, Individual)
        mutate = glyph.application.MutateFactory.create(args, Individual)
        select = glyph.application.SelectFactory.create(args)
        create_method = glyph.application.CreateFactory.create(args, Individual)

        ns = NullSpace(zero=args.constraints_zero, constant=args.constraints_constant, infty=args.constraints_infty)
        mate, mutate, Individual.create = apply_constraints([mate, mutate, Individual.create], constraints=build_constraints(ns))

        ndmate = partial(glyph.gp.breeding.nd_crossover, cx1d=mate)
        ndmutate = partial(glyph.gp.breeding.nd_mutation, mut1d=mutate)
        ndcreate = lambda size: [NDTree(create_method(args.ndim)) for _ in range(size)]
        NDTree.create_population = ndcreate
        algorithm_factory = partial(glyph.application.AlgorithmFactory.create, args, ndmate, ndmutate, select, ndcreate)
        assessment_runner = RemoteAssessmentRunner(send, recv, method=args.const_opt_method, options=args.options,
                                                   consider_complexity=args.consider_complexity, caching=args.caching, persistent_caching=args.persistent_caching,
                                                   simplify=args.simplify, chunk_size=args.chunk_size)
        gp_runner = glyph.application.GPRunner(NDTree, algorithm_factory, assessment_runner, callbacks=callbacks)
        app = RemoteApp(args, gp_runner, args.checkpoint_file)
    print_params(logger.info, vars(args))
    return app, args


def main():
    app, args = make_remote_app()

    bc = break_condition(ttl=args.ttl, target=args.target, max_iter=args.max_iter_total, error_index=0)
    app.run(break_condition=bc)

if __name__ == "__main__":
    main()



# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL

from .algorithms import *
from .algorithms import all_algorithms
from .breeding import all_mutations, all_crossover
from .constraints import NullSpace, apply_constraints, build_constraints
from .individual import AExpressionTree, sympy_primitive_set, sympy_phenotype, numpy_primitive_set, numpy_phenotype



# Copyright: 2017, Markus Abel, Julien Gout, Markus Quade
# Licence: LGPL

import os
import sys
import logging
import logging.config
import yaml

# module name is in conflict with stdlib and can cause unwanted monkey patching


def load_config(config_file, placeholders={}, default_level=logging.INFO):
    """Load logging configuration from .yaml file."""
    if not sys.warnoptions:
        # Route warnings through python logging
        logging.captureWarnings(True)
    if os.path.exists(config_file):
        with open(config_file, 'rt') as f:
            content = f.read().format(**placeholders)
        config = yaml.load(content)
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=default_level)


def log_level(verbosity):
    """Convert numeric verbosity to logging log levels."""
    level = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
    verbosity = len(level) - 1 if verbosity >= len(level) else verbosity
    return level[verbosity]


def print_dict(p_func, d):
    """
    Pretty print a dictionary

    :param p_func: printer to use (print or logging)
    :type d: dict
    """
    for k, v in sorted(d.items()):
        p_func('{} = {}'.format(k, v))


def print_params(p_func, gp_config):
    """Pretty print a glyph app config"""
    print_dict(p_func, gp_config)
    p_func('')



from setuptools import setup, find_packages

version = "0.3.1"


def read(fname, split=True):
    with open(fname, 'r') as f:
        content = f.read()
    return content.split('\n') if split else content


setup(
    name='pyglyph',
    version=version,
    author='Markus Abel, Julien Gout, Markus Quade',
    author_email='markus.quade@ambrosys.de',
    description="Symbolic regression tools.",
    long_description=read('README.rst', split=False),
    license='LGPL',
    keywords='complex systems, control, machine learning, genetic programming',
    url='https://www.github.com/ambrosys/glyph',
    packages=find_packages(exclude=["tests", "doc", "examples"]),
    install_requires=read('requirements-to-freeze.txt'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    entry_points={
        'console_scripts': [
            'glyph-remote = glyph.cli.glyph_remote:main'
        ]
    },
)
